"""
🧿 pupyl

Pupyl is a really fast image search library which you
can index your own (millions of) images and find similar
images in milliseconds.
"""
__version__ = 'v0.9.9'


import os
import json
import concurrent.futures

from pupyl.embeddings.features import Extractors, Characteristics
from pupyl.storage.database import ImageDatabase
from pupyl.indexer.facets import Index


class PupylImageSearch:
    """
    Encapsulates every aspect of pupyl, from feature extraction
    to indexing and image database.
    """
    def __init__(
            self,
            data_dir,
            **kwargs
    ):
        self._data_dir = data_dir
        self._index_config_path = os.path.join(self._data_dir, 'index.json')

        configurations = self._index_configuration('r')

        if configurations:
            self._import_images = configurations['import_images']
            self._characteristic = Characteristics.by_name(
                configurations['characteristic']
            )
        else:
            import_images = kwargs.get('import_images')
            characteristic = kwargs.get('characteristic')

            if import_images:
                self._import_images = import_images
            else:
                self._import_images = True

            if characteristic:
                self._characteristic = characteristic
            else:
                self._characteristic = Characteristics.\
                    LIGHTWEIGHT_REGULAR_PRECISION

        self.image_database = ImageDatabase(
            import_images=self._import_images,
            data_dir=self._data_dir
        )

    def _index_configuration(self, mode):
        """
        Load or save an index configuration file, if exists.

        Parameters
        ----------
        mode (values: ('r', 'w')): str
            Defines which mode should be used over configuration
            file. 'r' is for file reading, 'w' for writing.
        """
        try:
            with open(self._index_config_path, mode) as config_file:
                if mode == 'r':

                    return json.load(config_file)

                if mode == 'w':

                    configurations = {
                        'import_images': self._import_images,
                        'characteristic': self._characteristic.name
                    }

                    json.dump(configurations, config_file)

                return True
        except FileNotFoundError:
            return False

    def index(self, uri):
        """
        Performs image indexing.

        Parameters
        ----------
        uri: str
            Directory or file, or http(s) location.
        """
        with Extractors(
                characteristics=self._characteristic
            ) as extractor, Index(
                extractor.output_shape,
                data_dir=self._data_dir
            ) as index, concurrent.futures.ThreadPoolExecutor() \
                as executor:

            self._index_configuration('w')

            futures = {
                executor.submit(
                    extractor.save_tensor,
                    extractor.extract,
                    uri_from_file,
                    self.image_database.mount_file_name(rank, 'npy')
                ): {'id': rank, 'uri': uri_from_file}
                for rank, uri_from_file in enumerate(extractor.scan(uri))
            }

            for future in self.image_database.progress(
                concurrent.futures.as_completed(futures)
            ):

                self.image_database.insert(len(index), futures[future]['uri'])

                features_tensor_name = self.image_database.mount_file_name(
                    futures[future]['id'],
                    'npy'
                )

                index.append(extractor.load_tensor(features_tensor_name))

                os.remove(features_tensor_name)

    def search(self, query, top=4):
        """
        Executes the search for a created database

        Parameters
        ----------
        query: str
            URI of a image to query

        top (optional)(default: 4): int
            How many results should be returned.
        """
        with Extractors(characteristics=self._characteristic) as extractor:
            with Index(
                extractor.output_shape,
                data_dir=self._data_dir
            ) as index:
                for result in index.search(
                    extractor.extract(query),
                    results=top
                ):
                    yield result



"""
file_types Module.

Operations of file and mime type discovery, based on file content.
"""
from enum import Enum

from pupyl.duplex import exceptions


class FileHeaderHex(Enum):
    """Describes header hexadecimals for every supported file types."""

    JPG = 'ffd8ff'
    GIF = '474946'
    PNG = '89504e'
    TIF = '49492a'
    GZP = '1f8b08'
    BZ2 = '425a68'
    ZIP = '504b03'
    LXZ = 'fd377a'


class FileMimeTypes(Enum):
    """Describes the mime types associated with underlying files."""

    JPG = 'image/jpeg'
    GIF = 'image/gif'
    PNG = 'image/png'
    TIF = 'image/tiff'
    GZP = 'application/gzip'
    BZ2 = 'application/x-bzip2'
    ZIP = 'application/zip'
    LXZ = 'application/x-xz'


class FileType:
    """Operations of file type discovery over file content."""

    @staticmethod
    def to_hex(bytess):
        """
        Convert a byte string to its hexadecimal representation.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        Returns
        -------
        str:
            With the hexadecimal representation
        """
        return bytess.hex()

    @classmethod
    def header(cls, bytess):
        """
        Return only the file header or the first three bytes.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        Returns
        -------
        str:
            With the header hexadecimal representation
        """
        return cls.to_hex(bytess[:3])

    @classmethod
    def guess_file_type(cls, bytess, mimetype=False):
        """
        Heuristically discover the file type for the underlying path.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        mimetype (optional): bool
            If the mimetype should be returned or not

        Returns
        -------
        str:
            With inferred type
        """
        try:
            header = cls.header(bytess)
        except TypeError:
            raise exceptions.FileTypeNotSupportedYet

        for hexa in FileHeaderHex:
            if hexa.value == header:
                if mimetype:

                    return hexa, FileMimeTypes.__members__.get(hexa.name).value

                return hexa.name

        raise exceptions.FileTypeNotSupportedYet

    @classmethod
    def is_image(cls, bytess):
        """
        Return if the read file is an image or not.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        Returns
        -------
        bool:
            Describing if the inferred type is image
        """
        try:
            return cls.guess_file_type(
                bytess,
                mimetype=True
                )[1].split('/')[0] == 'image'
        except exceptions.FileTypeNotSupportedYet:
            return False



"""
file_io Module.

Operations over files, introspection and more.
"""
import os
from urllib.request import urlopen, urlparse
from enum import Enum, auto
import mimetypes
import tempfile
import uuid
import csv
import gzip
import zipfile
import bz2
import lzma
import tarfile
from datetime import datetime
from itertools import cycle
from io import BytesIO

import termcolor

from pupyl.duplex.file_types import FileType
from pupyl.duplex.exceptions import FileTypeNotSupportedYet, \
    FileScanNotPossible
from pupyl.addendum.operators import intmul


class Protocols(Enum):
    """Defines several possible protocols to be discovered."""

    UNKNOWN = auto()
    HTTP = auto()
    FILE = auto()


class FileIO(FileType):
    """Operations over files."""

    @staticmethod
    def _get_url(url):
        """
        Load a file from a remote (http(s)) location.

        Parameters
        ----------
        url: str
            The URL where the image are stored

        Returns
        -------
        bytes:
            With image binary information
        """
        with urlopen(url) as ffile:
            return ffile.read()

    @staticmethod
    def _get_local(path):
        """
        Load a local file.

        Parameters
        ----------
        path: str
            Local which the file is saved

        Returns
        -------
        bytes:
            With file binary information
        """
        with open(path, 'rb') as ffile:
            return ffile.read()

    @classmethod
    def get(cls, uri):
        """
        Load a file from specified location.

        Parameters
        ----------
        uri: str
            Location where the file are stored.

        Returns
        -------
        bytes or Enum:
            If successful, returns the image bytes,
            or an Enum describing format was not recognized
        """
        if cls._infer_protocol(uri) is Protocols.FILE:
            return cls._get_local(uri)

        if cls._infer_protocol(uri) is Protocols.HTTP:
            return cls._get_url(uri)

        return Protocols.UNKNOWN

    @classmethod
    def get_metadata(cls, uri):
        """
        Return underlying file metadata

        Parameters
        ----------
        uri: str
            Location where the file are stored.

        Returns
        -------
        dict:
            Describing several file metadata
        """
        if cls._infer_protocol(uri) is Protocols.FILE:
            file_statistics = os.stat(uri)

            original_path, original_file_name = os.path.split(uri)
            original_file_size = file_statistics.st_size // 2 ** 10
            original_access_time = cls.timestamp_to_iso8601(
                file_statistics.st_atime
                )

        if cls._infer_protocol(uri) is Protocols.HTTP:
            parsed_url = urlparse(uri)

            with urlopen(uri) as ffile:
                file_statistics = ffile.info()

            original_path, original_file_name = os.path.split(
                f'{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}'
            )

            try:
                measured_size = int(
                    file_statistics.get_all(
                        'Content-Length'
                    )[0]
                )
            except TypeError:
                with urlopen(uri) as ffile:
                    measured_size = len(ffile.read())

            original_file_size = measured_size // (2 ** 10)
            original_access_time = file_statistics.get_all('Date')[0]

        return {
            'original_file_name': original_file_name,
            'original_path': original_path,
            'original_file_size': f'{original_file_size}K',
            'original_access_time': original_access_time
        }

    @staticmethod
    def timestamp_to_iso8601(timestamp):
        """
        Convert a Unix epoch integer formatted date to ISO8601 format.
        The converted date is in UTC (GMT-0).

        Parameters
        ----------
        timestamp: int
            With a integer timestamp (seconds after zero hour of 1970)
        """
        return datetime.utcfromtimestamp(timestamp).\
            strftime('%Y-%m-%dT%H:%M:%S')

    @staticmethod
    def _infer_protocol(uri):
        """
        Discover the protocol which the passed uri may pertain.

        Parameters
        ----------
        uri: str
            URI that describes the image location

        Returns
        -------
        Enum:
            Referencing the discovered protocol
        """
        if urlparse(uri).scheme.startswith('http'):
            return Protocols.HTTP

        if uri.startswith('file') or os.path.exists(uri):
            return Protocols.FILE

        return Protocols.UNKNOWN

    def infer_file_type_from_uri(self, uri, mimetype=False):
        """
        Infer the file type from an uri.

        Parameters
        ----------
        uri: str
            With file to be analyzed

        mimetype (optional): bool
            If must be returns as mime type or not

        Returns
        -------
        str:
            If described file type
        """
        file_bytes = self.get(uri)

        try:
            if file_bytes is Protocols.UNKNOWN:
                raise FileTypeNotSupportedYet

            return self.guess_file_type(file_bytes, mimetype=mimetype)
        except FileTypeNotSupportedYet as file_type_not_supported_yet:
            file_type = mimetypes.guess_type(uri)[0]

            if file_type:
                return file_type if mimetype else \
                    file_type.split('/')[1].upper()

            raise file_type_not_supported_yet

    @classmethod
    def scan_csv(cls, uri):
        """
        Scanner for CSV formatted text files.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        csv_string = cls.get(uri)

        if csv_string is Protocols.UNKNOWN:
            raise FileTypeNotSupportedYet

        csv_string = csv_string.decode('utf-8').splitlines()

        reader = csv.reader(csv_string)

        for row in reader:
            yield row[0]

    @classmethod
    def scan_csv_gzip(cls, uri):
        """
        Scanner for CSV formatted text files, compressed with Gzip algorithm.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        file_bytes = cls.get(uri)

        gzip_file = gzip.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in gzip_file:
            yield row.replace('\n', '')

    @staticmethod
    def safe_temp_file(**kwargs):
        """
        Create a secure temporary file name,
        which means a file with an unique name.

        If a file with the same name is found,
        it's deleted a before generating a new unique name.

        Parameters
        ----------
        file_name (optional) (keyword argument): str
            Defining a temporary file to assert.
        """
        temp_dir = tempfile.gettempdir()

        if kwargs.get('file_name'):
            file_name = os.path.join(temp_dir, kwargs.get('file_name'))
        else:
            file_name = os.path.join(temp_dir, str(uuid.uuid4()))

        if os.path.exists(file_name):
            os.remove(file_name)

        return file_name

    @classmethod
    def scan_csv_bzip2(cls, uri):
        """
        Scanner for CSV formatted text files, compressed with Bzip2 algorithm.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        file_bytes = cls.get(uri)

        bz2_file = bz2.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in bz2_file:
            yield row.replace('\n', '')

    @classmethod
    def scan_csv_zip(cls, uri):
        """
        Scanner for CSV formatted text files, compressed with Zip algorithm.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        file_bytes = BytesIO(cls.get(uri))

        for ffile in zipfile.ZipFile(file_bytes).namelist():
            with zipfile.ZipFile(file_bytes).open(ffile) as ufile:
                for row in ufile:
                    yield row.decode('utf-8').replace('\n', '')

    @classmethod
    def scan_csv_xz(cls, uri):
        """
        Scanner for CSV formatted text files, compressed with Lzma algorithm.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        file_bytes = cls.get(uri)

        xz_file = lzma.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in xz_file:
            yield row.replace('\n', '')

    def scan(self, uri):
        """
        Return a validated uri, resolving several cases.

        It also choose the best discovery method.

        Parameters
        ----------
        uri: str
            A file or directory to scan

        Returns
        -------
        generator of str:
            With the discovery file paths
        """
        try:
            inferred_type = self.infer_file_type_from_uri(uri)

            if inferred_type in ('CSV', 'PLAIN'):
                for line in self.scan_csv(uri):
                    yield line

            elif inferred_type == 'GZP':
                for line in self.scan_csv_gzip(uri):
                    yield line

            elif inferred_type == 'ZIP':
                for line in self.scan_csv_zip(uri):
                    yield line

            elif inferred_type == 'BZ2':
                for line in self.scan_csv_bzip2(uri):
                    yield line

            elif inferred_type == 'LXZ':
                for line in self.scan_csv_xz(uri):
                    yield line

            else:
                raise FileScanNotPossible
        except IsADirectoryError:
            for root, _, files in os.walk(uri):
                for ffile in files:
                    yield os.path.join(root, ffile)

    @staticmethod
    def progress(iterable, precise=False):
        """
        Utility method to interface process progress bar with users.
        It supports two way of unpacking the iterable, throughout `precise`
        parameters. If `precise` is set to `False` (which is the default),
        the parameter `iterable` will be unpacked as is. This leads to an
        imprecise rolling of items (in other words, the method doesn't know
        apriori the total number of elements in `iterable`). Otherwise, if
        `precise` is set to `True`, an `iterable` which is not unpacked (like
        and `generator`) will be first unrolled, which is much slower in some
        cases, but leads to a precise progress bar.

        Parameters
        ----------
        iterable: iter
            Objects which supports iteration.

        precise (optional)(default: False): bool
            If the progress should be precise
            (with actual percentage of completion) or just an
            interface during process running.
        """
        if precise:
            try:
                count = len(iterable)
            except TypeError:
                iterable = [*iterable]
                count = len(iterable)

            print(
                termcolor.colored(
                    'Processing, please wait.',
                    color='green',
                    attrs=['bold']
                    )
                )

            for index, value in enumerate(iterable):
                try:
                    terminal_half_columns = os.get_terminal_size().columns // 4
                except OSError:
                    terminal_half_columns = 64

                percentage = (index + 1) / count

                columns_to_draw = percentage << intmul >> terminal_half_columns

                print(
                    ' ' + ('🟦' * columns_to_draw),
                    f'{percentage << intmul >> 100}%',
                    end='\r'
                )

                yield value
        else:
            clocks = '🕛🕐🕑🕒🕓🕔🕕🕖🕗🕘🕙🕚'

            for index, value_clock in enumerate(
                    zip(iterable, cycle(clocks))
            ):
                print(
                    ' ' + value_clock[1],
                    f' Processed {index + 1} items.',
                    end='\r'
                )

                yield value_clock[0]

    @staticmethod
    def resolve_path_end(path):
        """
        Removes directory separators from the end of
        some path (if exists).

        Parameters
        ----------
        path: str
            Complete path to be analyzed.
        """
        if path[-1] == os.path.sep:
            path = list(path)
            _ = path.pop()

            return ''.join(path)

        return path

    def dump(self, data_dir, output_dir):
        """
        Read an entire database tree, compress and export it.

        Parameters
        ----------
        data_dir: str
            The directory containing all database assets.

        output_dir: str
            Location where to save the export file.
        """
        data_dir = self.resolve_path_end(data_dir)
        base_name = os.path.basename(data_dir)
        dump_name = '.'.join((base_name, 'pupyl'))
        dump_path = os.path.join(output_dir, dump_name)

        with tarfile.open(dump_path, 'w:xz') as tar:
            for path in self.progress(self.scan(data_dir)):
                path_split = path.split(os.path.sep)
                archive_name = os.path.join(
                    *path_split[path_split.index(base_name):]
                )

                tar.add(path, arcname=archive_name)

    @staticmethod
    def bind(dump_file, output_dir):
        """
        Read a packaged database and import it.

        Parameters
        ----------
        dump_file: str
            The directory containing all database assets.

        output_dir: str
            Location where to save the export file.
        """
        with tarfile.open(dump_file) as dfile:
            dfile.extractall(output_dir)



"""
database module.

Operations and storage for images.
"""
import os
import json

from pupyl.duplex.image import ImageIO


class ImageDatabase(ImageIO):
    """Handling of storage and IO operations over images."""

    def __init__(self, import_images, data_dir=None, **kwargs):
        """
        Image storage and operations.

        Parameters
        ----------
        import_images: bool
            If the images must be copied to internal database or not.

        data_dir (optional) (default=self.safe_temp_file()): str
            Location to save the image storage files

        **bucket_size (optional) (default=1000): int
            Define the size of file buckets. In other words, how much
            files to save on a directory.

        **image_size (optional) (default=(800, 600)): tuple
            Define the dimensions of saved image. Only has some effect
            if import_images is True.
        """
        self._import_images = import_images

        if data_dir:
            self._data_dir = os.path.normpath(data_dir)
        else:
            self._data_dir = self.safe_temp_file()

        if kwargs.get('bucket_size'):
            self._bucket_size = kwargs['bucket_size']
        else:
            self._bucket_size = 1000

        if self._import_images:
            if kwargs.get('image_size'):
                self._image_size = kwargs['image_size']
            else:
                self._image_size = (800, 600)

        os.makedirs(self._data_dir, exist_ok=True)

    def __getitem__(self, position):
        """Return the item at index."""
        return self.load_image_metadata(position)

    def __len__(self):
        """Return how many items are indexed."""
        return len([*self.list_images()])

    @property
    def import_images(self):
        """Getter for import_images property."""
        return self._import_images

    @import_images.setter
    def import_images(self, import_it):
        """Setter for import_images property."""
        self._import_images = import_it

    @property
    def bucket_size(self):
        """Getter for bucket_size property."""
        return self._bucket_size

    @bucket_size.setter
    def bucket_size(self, size):
        """Setter for bucket_size property."""
        self._bucket_size = size

    @property
    def image_size(self):
        """Getter for image_size property."""
        if self._import_images:
            return self._image_size

        return (800, 600)

    @image_size.setter
    def image_size(self, dimensions):
        if self._import_images:
            self._image_size = dimensions

    def what_bucket(self, index):
        """
        Discover in what bucket the file should be saved.

        Parameters
        ----------
        index: int
            The index number which file will be referenced.

        Returns
        -------
        int:
            With the chosen bucket
        """
        return index // self._bucket_size

    def mount_file_name(self, index, extension):
        """
        Create the name of the posterior saved file.

        Parameters
        ----------
        index: int
            The indexer id associated with the file

        extension: str
            Describing the file extension
        """
        return os.path.join(
            self._data_dir,
            str(self.what_bucket(index)),
            f'{index}.{extension}'
        )

    def load_image_metadata(self, index, **kwargs):
        """
        Return the metadata inside image file metadata.

        Parameters
        ----------
        index: int
            Related to metadata file stored.

        **filtered (optional): iterable
            Describing which fields to filter (or select) for return.

        Returns
        -------
        dict
            Containing the parsed json file.
        """
        result_file_name = self.mount_file_name(index, 'json')

        try:
            with open(result_file_name) as json_file:
                metadata = json.load(json_file)

                if kwargs.get('filtered'):
                    metadata = {
                        key: value
                        for key, value in metadata.items()
                        if key in kwargs['filtered']
                    }

                return metadata

        except FileNotFoundError:
            raise IndexError

    def save_image_metadata(self, index, uri):
        """
        Store image metadata.

        Parameters
        ----------
        index: int
            The index related to the image.

        uri: str
            Local where image is stored.
        """
        bytess = self.get(uri)
        result_file_name = self.mount_file_name(index, 'json')

        os.makedirs(os.path.dirname(result_file_name), exist_ok=True)

        if self.is_image(bytess):
            with open(result_file_name, 'w') as json_file:
                metadata = self.get_metadata(uri)
                metadata['id'] = index

                json.dump(metadata, json_file)

    def insert(self, index, uri):
        """
        Insert an image to database.

        Parameters
        ----------
        index: int
            The index number attributed to the image.

        uri: str
            Where the original file are located
        """
        if self._import_images:
            self.save_image(
                self.mount_file_name(index, 'jpg'),
                self.compress(
                    self.size(uri, new_size=self._image_size, keep_aspect=True)
                )
            )

        self.save_image_metadata(index, uri)

    def list_images(self, return_index=False, top=None):
        """
        Return all images in current database.

        Parameters
        ----------
        return_index (optional)(default: False): bool
            If the method should also return the file index inside database.

        top (optional): int
            How many pictures from image database should be listed. Not setting
            this parameter (which means not referencing it or setting it
            to zero or below) will return all images in the database.
        """
        if top:
            counter = 0

        for root, _, files in os.walk(self._data_dir):

            for ffile in files:

                ffile = os.path.join(root, ffile)

                with open(ffile, 'rb') as t_file:
                    if self.is_image(t_file.read()):

                        if top:
                            counter += 1

                            if counter > top:
                                return

                        if return_index:
                            yield int(
                                os.path.splitext(
                                    os.path.split(ffile)[1]
                                )[0]
                            ), ffile
                        else:
                            yield ffile

    def load_image(self, index):
        """Return the image data at specified index."""
        return self.get_image(self.mount_file_name(index, 'jpg'))



"""
Exceptions regarding embeddings module operations
"""


class Error(Exception):
    """
    Base exception class
    """


class UnknownCharacteristics(Error):
    """
    Raised when a not known network characteristic is passed
    """



"""
Image module.

Performs multiple operations over images, like resizing,
loading and so on.
"""
from base64 import b64encode

import numpy
import tensorflow
from tensorflow import io as io_ops
from tensorflow import image as image_ops

from pupyl.duplex.file_io import FileIO
from pupyl.duplex.exceptions import FileIsNotImage


tensorflow.autograph.set_verbosity(0)


class ImageIO(FileIO):
    """Operations of read and write over images."""

    @staticmethod
    def encoded_to_compressed_tensor(encoded):
        """
        Transform a binary encoded string image to its compressed tensor.

        Parameters
        ----------
        encoded: bytes
            The binary representation of the image

        Returns
        -------
        numpy.ndarray
        """
        return numpy.array(memoryview(io_ops.decode_raw(encoded, 'uint8')))

    @staticmethod
    def encoded_to_tensor(encoded):
        """
        Transform a binary encoded string image to tensor.

        Parameters
        ----------
        encoded: bytes
            The binary representation of the image

        Returns
        -------
        numpy.ndarray
            Describing the compressed image tensor
        """
        return numpy.array(memoryview(io_ops.decode_image(encoded)))

    @classmethod
    def get_image(cls, uri, as_tensor=False):
        """
        Load a image from specified location.

        Parameters
        ----------
        uri: str
            Location where the image are stored

        as_tensor (optional): bool
            Default: False

            If the image should be converted to its tensor representation.
            Default to False, which returns images to byte representation

        Returns
        -------
        bytes or numpy.ndarray
            Respectively returns the image as bytes or as tensor
            representation
        """
        bytess = cls.get(uri)

        if cls.is_image(bytess):
            if as_tensor:
                return cls.encoded_to_tensor(bytess)

            return bytess

        raise FileIsNotImage

    @classmethod
    def get_image_base64(cls, uri):
        """
        Load an image as a base64 encoded string.

        Parameters
        ----------
        uri: str
            Location where the image is stored.
        """
        return b64encode(cls.get_image(uri, as_tensor=False))

    @classmethod
    def get_image_bytes_to_base64(cls, image_bytes):
        """
        Return the image representation in base64, from its
        bytes representation.

        Parameters
        ----------
        image_bytes: bytes
            Bytes representation of some image.
        """
        if cls.is_image(image_bytes):
            return b64encode(image_bytes)

        raise FileIsNotImage

    @staticmethod
    def save_image(path, bytess):
        """
        Save an image to defined path.

        Parameters
        ----------
        path: str
            Local to save the image.

        bytess: bytes
            Bytes representing an image.
        """
        io_ops.write_file(path, bytess)

    @classmethod
    def size(cls, uri, new_size=None, keep_aspect=False):
        """
        Return the current image dimensions or resize it.

        Parameters
        ----------
        uri: str
            Description of where the image are located

        new_size (optional): tuple
            The new intended dimension of the image

        keep_aspect (optional) (default=False): bool
            If the image proportions should be preserved
            or not.

        Returns
        -------
        tuple:
            Current image dimensions

        If new_size is True:
            numpy.ndarray:
                A resized image, if new_size parameter
                was passed through
        """
        if new_size:
            return numpy.array(
                memoryview(
                    image_ops.resize(
                        cls.get_image(uri, as_tensor=True),
                        new_size,
                        method=image_ops.ResizeMethod.NEAREST_NEIGHBOR,
                        preserve_aspect_ratio=keep_aspect
                    )
                )
            )

        return cls.get_image(uri, as_tensor=True).shape[:2]

    @classmethod
    def compress(cls, tensor, as_tensor=False):
        """
        Compress the tensor using JPEG algorithm.

        Parameters
        ----------
        tensor: numpy.ndarray
            A tensor representing an image

        as_tensor (optional) (default=False): bool
            If the new compressed JPEG image should be
            returned as a numpy.ndarray

        Returns
        -------
        numpy.ndarray
            An encoded image, in bytes or numpy.ndarray
        """
        compressed = image_ops.encode_jpeg(
            tensor,
            quality=80,
            progressive=True,
            optimize_size=True,
            chroma_downsampling=True
        )

        if as_tensor:
            return cls.encoded_to_tensor(compressed)

        return compressed



"""
interface Module.

Create a web interface to query images and see results
based on indexed images on database.
"""
import os
import webbrowser
from http.server import SimpleHTTPRequestHandler
import socketserver
from urllib.parse import urlparse, parse_qs
import termcolor

from pupyl.search import PupylImageSearch


STATIC_FOLDER = os.path.abspath(os.path.join('web', 'static'))
TEMPLATE_FILE = os.path.join(STATIC_FOLDER, 'template.html')


def serve(data_dir, port=8080):
    """
    Start the web server.

    Parameters
    ----------
    port (optional)(default: 8080): int
        Defines the network port which the web server
        will start listening.
    """

    pupyl_image_search = PupylImageSearch(data_dir)

    class RequestHandler(SimpleHTTPRequestHandler):
        """A web request handler."""

        _data_dir = data_dir

        def __init__(self, request, client_address, server):
            SimpleHTTPRequestHandler.__init__(
                self, request, client_address, server
            )

        def do_GET(self):
            """Handler for GET request methods."""

            query_image = None

            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()

            query_string = parse_qs(urlparse(self.path).query)

            query_list = query_string.get('uri', None)

            image_tags = self.images(query_list)

            with open(TEMPLATE_FILE) as template:
                html_template = template.read()

            if query_list:
                query_image = '<img class="img-thumbnail" ' + \
                    f'src="{query_list[0]}">' + \
                    '<figcaption class="figure-caption">' + \
                    'Query image used in the search.</figcaption>'

            self.wfile.write(
                bytes(
                    html_template.format(
                        images=image_tags,
                        query=query_image if query_image else ''),
                    'utf-8'
                )
            )

        @staticmethod
        def filter_metadata(index):
            """
            Return a filtered metadata information.

            Parameters
            ----------
            index: int
                Index number of image

            Returns
            -------
            dict:
                With a pre-filtered metadata.
            """
            metadata = pupyl_image_search.image_database.\
                load_image_metadata(
                    index, filtered=(
                        'original_file_name',
                        'original_file_size'
                    )
                )

            return ', '.join(map(str, metadata.values()))

        def images(self, query_uri=None, top=None):
            """
            Return image tags from database.

            Parameters
            ----------
            query_uri (optional): str
                Location where the query image is stored.

            top (optional)(default: 24): int
                How many results should be returned from some search request.
            """

            image_tags = ''
            img_src = '<figure class="figure">' + \
                '<img class="img-fluid border rounded-circle"' + \
                'src="data:image/jpg;base64, {image_b64}" ' + \
                'alt="&#129535; Pupyl"><figcaption class="figure-caption">' + \
                '{figure_caption}</figcaption></figure>'

            top = top if top else 24

            if query_uri:
                query_uri = query_uri[0]

                for result in pupyl_image_search.search(query_uri, top=top):
                    result = int(result)

                    image = pupyl_image_search.image_database.\
                        get_image_bytes_to_base64(
                            pupyl_image_search.image_database.
                            load_image(result)
                        ).decode('utf-8')

                    filtered_metadata = self.filter_metadata(result)

                    image_tags += img_src.format(
                        image_b64=image,
                        figure_caption=filtered_metadata
                    )

                return image_tags

            for index, image in pupyl_image_search.image_database.list_images(
                    return_index=True,
                    top=9
            ):
                image_base64 = pupyl_image_search.image_database.\
                    get_image_base64(
                        image
                    ).decode('utf-8')

                filtered_metadata = self.filter_metadata(index)

                image_tags += img_src.format(
                    image_b64=image_base64,
                    figure_caption=filtered_metadata
                )

            return image_tags

    if not port:
        port = 8080

    try:
        with socketserver.TCPServer(('', port), RequestHandler) as httpd:
            print(
                termcolor.colored(
                    f'Server listening on port {port}.',
                    color='green',
                    attrs=['bold']
                )
            )

            webbrowser.open_new_tab(f'http://localhost:{port}')
            httpd.serve_forever()

    except OSError:
        print(
            termcolor.colored(
                f'Port {port} already in use. Trying {port + 1}...',
                color='red',
                attrs=['bold']
            )
        )

        serve(data_dir=data_dir, port=port + 1)



"""Factory for image feature extraction."""
import os
import warnings
from enum import Enum, auto
import termcolor

import numpy
import tensorflow
import tensorflow.keras.applications as networks
import tensorflow.keras.backend as backend

from pupyl.embeddings import exceptions
from pupyl.duplex.image import ImageIO


tensorflow.autograph.set_verbosity(0)


class Characteristics(Enum):
    """Describes high level characteristics of complex feature extractors."""

    # MobileNetV2
    LIGHTWEIGHT_REGULAR_PRECISION = auto()
    # DenseNet169
    MEDIUMWEIGHT_GOOD_PRECISION = auto()
    # NASNetLarge
    HEAVYWEIGHT_HUGE_PRECISION = auto()

    def by_name(name):
        """
        Return a characteristic by name.

        Parameters
        ----------
        name: str
            String representation of enumerator
        """
        return Characteristics[name]


class Extractors(ImageIO):
    """Pretrained CNNs for embeddings generation."""

    def __init__(self, characteristics):
        """
        Create embedding extractors.

        Parameters
        ----------
        characteristics: Enum
            Describing the intended characteristics to transform images
            into embeddings
        """
        self.acceleration_discovery()

        self._characteristics = characteristics

        self.converter, self.network = self._infer_network()
        self.image_input_shape = self.network.input_shape[1:3]
        self._features_output_shape = self.network.output_shape[1]

    def __enter__(self):
        """Open Extractors context."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close Extractors context."""
        del exc_type, exc_val, exc_tb

        backend.clear_session()

    @property
    def output_shape(self):
        """
        Getter for property output_shape.

        Describes the output shape of extracted features.
        """
        return self._features_output_shape

    @staticmethod
    def acceleration_discovery():
        """
        Perform a hardware processing acceleration discovery.

        Most GPUs supported (through CUDA), which results on
        faster embeddings extraction.
        """
        # If GPU(s) device(s) are found, avoid excessive VRAM consumption
        for gpu in tensorflow.config.experimental.list_physical_devices('GPU'):
            print(
                termcolor.colored(
                    'GPU found! Using it for embeddings '
                    'extraction acceleration.',
                    color='blue',
                    attrs=['bold']
                    )
                )

            try:
                tensorflow.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError:
                warnings.warn(
                    termcolor.colored(
                        'Trying to define memory growth for an '
                        'already initialized GPU',
                        color='blue',
                        attrs=['bold']
                        ),
                    ResourceWarning
                    )

    def _infer_network(self):
        """Translate a characteristic to a network architecture."""
        # Networks general configurations
        weights = 'imagenet'
        pooling = 'max'
        include_top = False
        input_shape = (224, 224, 3)
        input_shape_nasnet_large = (331, 331, 3)

        if self._characteristics is \
                Characteristics.LIGHTWEIGHT_REGULAR_PRECISION:
            return networks.mobilenet_v2.preprocess_input, \
                networks.mobilenet_v2.MobileNetV2(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=input_shape
                    )

        if self._characteristics is \
                Characteristics.MEDIUMWEIGHT_GOOD_PRECISION:
            return networks.densenet.preprocess_input, \
                networks.densenet.DenseNet169(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=input_shape
                    )

        if self._characteristics is \
                Characteristics.HEAVYWEIGHT_HUGE_PRECISION:
            return networks.nasnet.preprocess_input, \
                networks.nasnet.NASNetLarge(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    # Specific shape for NASNetLarge
                    input_shape=input_shape_nasnet_large
                    )

        raise exceptions.UnknownCharacteristics

    def preprocessor(self, uri):
        """
        Image preprocessing methods, suitable for posterior network ingestion.

        Parameters
        ----------
        image: numpy.ndarray
            Tensor which represents an image

        Returns
        -------
        numpy.ndarray
            Containing the processed image
        """
        return numpy.expand_dims(
            self.converter(self.size(uri, self.image_input_shape)),
            axis=0
            )

    def extract(self, uri):
        """
        Convert image uri to its embeddings.

        Parameters
        ----------
        uri: str
            Tensor which represents an image

        Returns
        -------
        numpy.ndarray
            1D tensor with extracted features
        """
        return self.network.predict(
            self.preprocessor(uri)
            ).ravel()

    @staticmethod
    def save_tensor(func_gen, uri, file_name):
        """
        Saves an arbitrary tensor to file.

        Parameters
        ----------
        func_gen: function
            The generator function of the tensor.

        uri: str
            Location of image to generate tensor.

        file_name: str
            Path with file name.
        """
        os.makedirs(os.path.dirname(file_name), exist_ok=True)

        numpy.save(
            file_name,
            func_gen(uri),
            allow_pickle=False,
            fix_imports=False
        )

    @staticmethod
    def load_tensor(file_name):
        """
        Saves an arbitrary tensor to file.

        Parameters
        ----------
        file_name: str
            The file name inside features database.
        """
        return numpy.load(
            file_name,
            mmap_mode='r',
            allow_pickle=False,
            fix_imports=False
        )



"""
facets Module.

Hyperspace indexing and operations.
"""
import os
from shutil import move

from annoy import AnnoyIndex

from pupyl.indexer.exceptions import FileIsNotAnIndex, IndexNotBuildYet, \
    NoDataDirForPermanentIndex, DataDirDefinedForVolatileIndex, \
    NullTensorError
from pupyl.addendum.operators import intmul
from pupyl.duplex.file_io import FileIO


class Index:
    """Procedures over multidimensional spaces."""

    def __init__(self, size, data_dir=None, trees=.001, volatile=False):
        """
        Indexing tensors operations and nearest neighbours search.

        Parameters
        ----------
        size: int
            Shape of unidimensional vectors which will be indexed

        data_dir: str
            Location where to load or save the index

        trees (optional): float
            Defines the number of trees to create based on the dataset
            size. Should be a number between 0 and 1.

        volatile (optional): bool
            If the index will be temporary or not.
        """
        self._position = -1
        self._size = size
        self._data_dir = data_dir
        self._trees = trees
        self._volatile = volatile

        if self._data_dir and not self._volatile:
            if os.path.isfile(self._data_dir):
                raise OSError('data_dir parameter is not a directory' +
                              ', is a file.')

            os.makedirs(self._data_dir, exist_ok=True)
            self._path = os.path.join(self._data_dir, self.index_name)
        elif not self._data_dir and not self._volatile:
            raise NoDataDirForPermanentIndex
        elif not self._data_dir and self._volatile:
            _temp_file = FileIO.safe_temp_file()
            self._data_dir = os.path.dirname(_temp_file)
            self._path = _temp_file

        else:
            raise DataDirDefinedForVolatileIndex

        if os.path.isfile(self._path):
            try:
                self.tree = AnnoyIndex(size, metric='angular')

                self.tree.load(self._path)

                self._is_new_index = False
            except OSError:
                raise FileIsNotAnIndex
        else:
            self.tree = AnnoyIndex(size, metric='angular')
            self._is_new_index = True

    @property
    def size(self):
        """Getter for property size."""
        return self._size

    @property
    def path(self):
        """Getter for property path."""
        return self._path

    @property
    def index_name(self):
        """Getter for property index_name."""
        return 'pupyl.index'

    @property
    def trees(self):
        """Getter for property trees."""
        return self._trees

    @property
    def volatile(self):
        """Getter for property volatile."""
        return self._volatile

    @trees.setter
    def trees(self, trees):
        """Setter for property trees."""
        self._trees = trees

    def __enter__(self):
        """Context opening index."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context closing index."""
        del exc_type, exc_val, exc_tb

        if self._is_new_index:
            self.tree.build(self.size << intmul >> self.trees)

            self.tree.save(self.path)

        self.tree.unload()

    def items(self):
        """Return the indexed items."""
        for item in range(self.tree.get_n_items()):
            yield item

    def values(self):
        """Return the indexed values."""
        for item in self.items():
            yield self.tree.get_item_vector(item)

    def items_values(self):
        """Return tuples with all items and values."""
        for item, value in zip(self.items(), self.values()):
            yield item, value

    def __getitem__(self, position):
        """Return item at index. Supports negative slicing."""
        if position >= 0:
            return self.tree.get_item_vector(position)

        return self.tree.get_item_vector(
            self.tree.get_n_items() - abs(position)
            )

    def refresh(self):
        """Update all information regarding index file."""
        self.tree.unload()
        self.tree.load(self.path)

    def append(self, tensor):
        """
        Insert a new tensor at the end of the index.
        Be advised that this operation is linear on index size ($O(n)$).

        Parameters
        ----------
        tensor: numpy.ndarray or list
            A vector to insert into index.
        """
        if sum(tensor) == 0.:
            raise NullTensorError

        if self._is_new_index:
            self.tree.add_item(len(self), tensor)
        else:
            with Index(self.size, volatile=True, trees=self.trees) as tmp_idx:
                for value in self.values():
                    tmp_idx.append(value)

                tmp_idx.append(tensor)

                _temp_file = tmp_idx.path

            move(_temp_file, self.path)

            self.refresh()

    def remove(self, position):
        """
        Remove the tensor at index from the database.
        Be advised that this operation is linear on index size ($O(n)$).

        Parameters
        ----------
        position: int
            The index which must be removed
        """
        if self._is_new_index:
            raise IndexNotBuildYet

        if position > len(self):
            raise IndexError

        with Index(self.size, volatile=True, trees=self.trees) as tmp_idx:
            shrink = False

            for item, value in self.items_values():
                if item == position:
                    shrink = True
                else:
                    if shrink:
                        item -= 1

                    tmp_idx.tree.add_item(item, value)

            _temp_file = tmp_idx.path

        move(_temp_file, self.path)

        self.refresh()

    def pop(self, position=None):
        """
        Pop-out the index at position, returning it.
        Be advised that this operation is linear on index size ($O(n)$).

        Parameters
        ----------
        position (optional) (default: last position): int
            Removes and returns the value at position.

        Returns
        ----------
        int:
            With the popped item.
        """
        if position:
            value = self[position]
        else:
            inverse_index = -1
            value = self[inverse_index]
            position = len(self) + inverse_index

        self.remove(position)

        return value

    def index(self, tensor):
        """
        Search for the first most similar image compared to the query.

        Parameters
        ----------
        tensor: numpy.ndarray or list
            A vector to insert into index.

        Returns
        ----------
        int:
            Describing the most similar resulting index.
        """
        return self.tree.get_nns_by_vector(tensor, n=1)[0]

    def search(self, tensor, results=16):
        """
        Search for the first most similar image compared to the query.

        Parameters
        ----------
        tensor: numpy.ndarray or list
            A vector to insert into index

        results: int
            How many results to return. If similar images are less than
            results, it exhausts and will be returned actual total.
        """
        for result in self.tree.get_nns_by_vector(tensor, n=results):
            yield result

    def __len__(self):
        """Return how many items are indexed."""
        return self.tree.get_n_items()

    def __iter__(self):
        """Return an iterable."""
        for value in self.values():
            yield value

    def __next__(self):
        """Iterate over the iterable."""
        self._position += 1

        all_values = list(self.values())

        if self._position < len(all_values):
            return all_values[self._position]

        raise StopIteration



"""
operators Module.

Overloaded and new python operators.
"""


class Infix:
    """Definitions for infix operators."""

    def __init__(self, method):
        """Infix operators."""
        self.method = method

    def __rlshift__(self, other):
        """Supporting reversed right hand operations."""
        return Infix(lambda x, self=self, other=other: self.method(other, x))

    def __rshift__(self, other):
        """Supporting reversed right hand operations."""
        return self.method(other)


# intmul
# Performs an integer multiplication.
# If result is less than or equal to zero, it bounds to one.
#

intmul = Infix(lambda x, y: 1 if int(x * y) <= 0 else int(x * y))



"""
Exceptions regarding embeddings module operations
"""


class Error(Exception):
    """
    Base exception class
    """


class FileIsNotAnIndex(Error):
    """
    Raised when a file is not recognized as a index file.
    """


class IndexNotBuildYet(Error):
    """
    Raised when a file is not recognized as a index file.
    """


class NoDataDirForPermanentIndex(Error):
    """
    Raised when a data directory isn't defined for a not volatile index.
    """


class DataDirDefinedForVolatileIndex(Error):
    """
    Raised when a data directory is defined for a volatile index.
    """


class NullTensorError(Error):
    """
    Raised when trying to insert a null tensor.
    """
